#include "calc.h"

extern struct gl_nodeN treeN[];
extern int nodesN;
extern int treeOrderN[];
extern int treePosN[];

extern float treeGainWeight[];
extern float treeLossWeight[];

/* -------------------------------------------------------------------------- */

/* do forward and backward recursion */
void gl_calc(struct gl_arguments ga) {

  int lca, lca_pos, kMin, kMax, m, i, j;

  if(ga.collectfile != NULL) {
    lca = gl_getLCA(&kMin, &kMax); /* label of root of subtree */
    gl_printTreePSN(lca, ga.psfile, ga.type);
    gl_printTreeN(ga.outfile);
    return;
  }


  //  int done[N_NODES];
  float **S=NULL, inf=9999., SS; /*, p, q, x;*/
  int k;/*kk, left, right, l, leaves[N_NODES];*/
  char * psfname=NULL;
  //  float TOL = 0.0001;


  /* print treeN, and the order and pos array */
  /* for(i=0; i<nodesN; i++) { */
  /*   j = treeOrderN[i]; /\* label *\/ */
  /*   m = treePosN[j]; /\* pos in tree *\/     */
  /*   printf("treeOrderN[%3d] =  %3d (label)\t",i,j); */
  /*   printf("treeN[%3d].pInP = %3d\t",m,treeN[m].pInP); */
  /*   printf("treePosN[%3d] = %3d (pos in tree)\t", j, m); */
  /*   printf("treeN[%3d].n = %3d (node.n)\tm:%3d",m,treeN[m].n,treeN[m].m); */
  /*   printf(" -- %s\n",treeN[m].o); */
  /* } */

  // gl_print
  /* gl_printTreeN(NULL); */
  lca = gl_getLCA(&kMin, &kMax); /* label of root of subtree */
  
  if(lca == -1) { printf("Your species are not in the tree\nExiting program."); return; }
  
  lca_pos = treePosN[lca];
  printf("\nLCA - root of subtree: %d pos: %d = %s\n", lca, lca_pos, treeN[lca_pos].o);
  printf("kMin: %d kMax: %d\n", kMin, kMax);
  
  /* get position in treeN of right-most last node */
  //r = treePosN[treeOrderN[nodesN-1]];
  
  /* allocate Score array k-> rows (# mirs), j -> cols (nodes, in preorder)*/
  S = (float **) calloc(kMax+1, sizeof(float *));
  for(i=0; i<=kMax; i++) {
    S[i] = (float *) calloc(nodesN+1, sizeof(float));
    for(j=0; j<nodesN; j++) {
      S[i][j] = -1.; /* preorder does not matter jet */
    }
  }
  
  /* initialize leaves */
  //  l = 0;
  for(j=0; j<nodesN; j++) {
    if(treeN[j].nc == 0) { /* this is a leaf node for it has no children */
      //      leaves[l] = j; l++;    /* remember order of leaves - do i need that?*/
      for(i=0; i<=kMax; i++) { /* check the number of miRNAs at this leaf */
	if(treeN[j].m == i) S[i][j] = 0.;
	else S[i][j] = inf;
      }
    } 
  }

  /* printf("\nScore array after initialization (before forward recursion):\n"); */
  /* gl_printS(S, kMax+1, nodesN); */

  int c,pos, cpos, mink;
  float minscore,score,sum;
  
  /* start forward recursion in lca to get optimal score */
  /* work on positions in tree, parse labels to positions before */
  lca_pos = treePosN[lca];
  j = lca_pos; /* j is position in treeN */
  while(S[kMax][lca_pos] == -1) {
    /* printf("FW: j=%d nc: %d treeN[%d].o=%s\n",j,treeN[j].nc, j,treeN[j].o); fflush(stdout); */
    if(treeN[j].nc == 0) { /* this is a leaf, that has already been scored */
      treeN[j].seenFW = 1;
      j = treePosN[(treeN[j].p)->n];
      goto nextWhileFW;
    }

    /* go through all child nodes, dig into those that have not been scored so far */
    for(i=0; i<treeN[j].nc; i++) {
      pos = treePosN[treeN[j].c[i]];
      /* if(S[1][pos] == -1) {  *//* child that has not been scored so far */
      if(treeN[pos].seenFW == -1) {
	j = pos;
	/* printf("go down to child to pos: %d\n",j); */
	goto nextWhileFW;
      }
    }

    /* all childnodes have been scored - check inner node */
    if(treeN[j].P == 0) { S[0][j] = 0; }
    else { S[0][j] = inf; }

    
    for(i=1; i<=kMax; i++) { /* for each row */
      sum=0;
      for(c=0; c<treeN[j].nc; c++) {
	/* child i */
	pos = treePosN[treeN[j].c[c]]; /* position of child in tree and in j of S[k][j]*/
	/* printf("position: pos = treePosN[%d] = %d\n",treeN[j].c[c],treePosN[treeN[j].c[c]]); */
	minscore = inf;
	score = inf;
	for(k=0; k<=kMax; k++) {
	  score = S[k][pos] + gl_delta(i,k,pos);
	  /* printf("score=%f + %f = %f (i: %d k: %d pos: %d)\n",S[k][pos],gl_delta(i,k,pos),score,i,k,pos); */
	  if(score < minscore) { minscore = score; }
	}
	sum += minscore;
      }
      S[i][j] = sum; 
      /* gl_printS(S, kMax+1, nodesN); */
    }
    treeN[j].seenFW = 1;
    
    /* END of scoring, go to parent node */
    if(treeN[j].p == NULL) { break; } /* stop if root of T is reached */
    j = treePosN[(treeN[j].p)->n]; 
    goto nextWhileFW;
    
  nextWhileFW:
    continue;
  
  }

  /* printf("Score array after forward recursion:\n"); */
  /* gl_printS(S, kMax+1, nodesN); */

  /* adjust gain and loss of complete gene family */
  int jP;
  for(j=0; j<nodesN; j++) {
    if((treeN[j].p) == NULL) continue;
    jP = treePosN[(treeN[j].p)->n];
    if(S[0][j] == 0 && S[0][jP] >= inf) {
      treeN[j].lossFam = 1;
    } else {
      treeN[j].lossFam = 0;
    }
    treeN[j].gainFam = 0;
  }

  /* start backtracing to label inner nodes */
  //  for(j=0; j<N_NODES; j++) done[j] = -1;
  //  printf("hello\n"); exit(0);  
  /* find optimal score for lca */
  i = 1;
  SS = inf;
  /* this variant results in first minimum! -> favours gain after lca */
  /* start from kMax to k=1 to favour loss after lca instead */
  //for(k=kMax; k>=0; k--) {
  for(k=0; k<=kMax; k++) { 
    if(S[k][lca_pos] <= SS) { 
      SS = S[k][lca_pos];
      i = k;
    }
  }
  fprintf(stdout,"Optimal score at LCA=S[%d][%d]: %.2f (# of miRNAs: %d)\n\n",i,lca_pos,SS,i);
  /* i carries score of inner node */
  /* add gain of whole family */
  if(treeN[lca_pos].p == NULL && i == 0) { // members of this family are not in tree
    treeN[lca_pos].gainFam = 0;
  } else {
     treeN[lca_pos].gainFam = 1;
  }

  j = lca_pos;
  k = kMax*inf; //kk = kMax*inf;
  
  /**** start bw recursion */

  /* initalize lca and done list for lca */
  //done[j] = 1;
  treeN[j].m = i;
  treeN[j].seenBW = 1;
  treeN[j].gain = i;
  treeN[j].loss = 0;
  int seenNodes = 1;
  /* initialize all nodes before lca and their seenBW with 0 and 1 */
  /* for(c=0; c<nodesN; c++) { */
  /*   if(treeOrderN[c] == treeN[lca_pos].n) { break; } */
  /*   if(treeOrderN[c]  */
  /*   lab = treeOrderN[c]; */
  /*   pos = treePosN[lab]; */
  /*   treeN[pos].seenBW = 1; */
  /*   treeN[pos].m = 0; */
  /* } */


    //  printf("mir at lca: %d\n",treeN[j].m); exit(0);
  for(c=0; c<nodesN; c++) {
    /* printf("p: %d l: %d m: %d seenBW: %d (%s)\n",c,treeN[c].n,treeN[c].m,treeN[c].seenBW,treeN[c].o); */
    if(treeN[c].seenBW == 1) { seenNodes++; } /* every leaf and all nodes outside of tree rooted at lca need not to be visited */
  }

  /* printf("dre: %d (%s) (seenBW: %d)\n",treeN[192].m,treeN[192].o,treeN[192].seenBW);     */
  while(seenNodes <= nodesN) {
    /* printf("j: %d i: %d (seenNodes: %d nodesN: %d) seenBW: %d\n",j,i, seenNodes, nodesN,treeN[j].seenBW); */
    /* gl_printTreeN(ga.outfile); */

    if(treeN[j].seenBW == 1) { /* already checked node */
      
      /* check if all children have been seen */
      for(c=0; c<treeN[j].nc; c++) {
	cpos = treePosN[treeN[j].c[c]];
	if(treeN[cpos].seenBW == -1) {
	  j = cpos;
	  i = treeN[j].m;
	  goto nextWhile;
	}
      }

      /* all children scored, go to parent */
      if(j == lca_pos) { i = 0; }
      else { i = (treeN[j].p)->m; }
      if(treeN[j].p) {
	//	j = (treeN[j].p)->pInT;
	j = treePosN[(treeN[j].p)->n];
	goto nextWhile;
      } else {
	//	printf("breaking\n");
	break;
      }
      
    } else { /* node to be checked */

      /* if not checked leave */
      if(treeN[j].nc == 0) {
	i = treeN[j].m;
      } else {
	 
	/* since I select only the minimum (over k) of S[k][j] in fw recursion,
	   I can do this here as well */
	/* yes, but include delta function!! */
	minscore = inf;
	mink = kMax;
	for(k=0; k<=kMax; k++) {
	  if(S[k][j] + gl_delta((treeN[j].p)->m,k,j) < minscore) {
	    minscore = S[k][j] +  gl_delta((treeN[j].p)->m,k,j);
	    mink = k;
	  }
	}
	i = mink;
      }
      /* m is difference of parent.m (NULL at lca -> -i) and current i (number of miRNAs) */
      if(j == lca_pos || treeN[j].p == NULL) { m = -i; }
      else { m = (treeN[j].p)->m - i; }

      /* printf("j:%d i:%d m: %d\n",j,i,m);  */

      /* adjust gain/loss */
      if(m < 0) { treeN[j].gain = abs(m); }
      else { treeN[j].loss = m; }
      
      //if(treeN[j].nc == 0) {
	treeN[j].m = i;
	//}
      
      treeN[j].seenBW = 1;
      
      seenNodes++;
      goto nextWhile;
      
      
    }
    
  nextWhile:
    continue;
  }
  /* printf("gmo: %d (%s)\n",treeN[189].m,treeN[189].o);     */
  if(ga.psfile == NULL) {
    psfname = (char *) calloc(strlen(ga.infile)+4, sizeof(char));
    strcpy(psfname, ga.infile);
    strcat(psfname, ".ePoPE.out.ps");
  } else {
    psfname = (char *) calloc(strlen(ga.psfile)+4, sizeof(char));
    strcpy(psfname, ga.psfile);
  }
  //  gl_printTreePS(lca, psfname);
  gl_printTreePSN(lca, psfname, ga.type);
  fprintf(stdout, "The tree is VISUALIZED in '%s'.\n", psfname);
  free(psfname);

  //  gl_printTree(ga.outfile);
  gl_printTreeN(ga.outfile);
  for(i=0; i<=kMax; i++) free(S[i]);
  free(S);

  return;

}

/* -------------------------------------------------------------------------- */

/* delta function that returns corresponding weight */
float gl_delta(int i, int k, int node) {

  float x, inf=9999., delta=0.;

  if(i==0 && k == 0) { delta = inf; }
  else {
    //x = fabs((float)(k - i));			
    x = (float)(k - i);
    if(x == 0) { delta = 0.; }
    else if(x < 0) {
      delta = fabs(x)*treeGainWeight[node];
    } else {
      delta = x*(1+treeLossWeight[node]);
    }
  }
  //  printf("delta = %f (%s) gainW: %f lossW: %f\n",delta,treeN[node].o,treeGainWeight[node],treeLossWeight[node]);
  return delta;
}

/* -------------------------------------------------------------------------- */

/* find index of node that is last common ancestor of current miRNA family
   assign min and max number of miRNAs in current alignment 
   returns label (!)
 */
int gl_getLCA(int *kMin, int *kMax) {

  int i, j, lab=0, pos=0, labFirst=-1, labLast=-1, lca=0;
  int posFirst=-1,posLast=-1; /* position of first and last sp. with gene members in preorder */
  /* these are the possible values for min and max number of miRNAs in the leafs */
  *kMin = 9999;
  *kMax = 0;

  /* find first (acc. to preorder) leaf with miRNAs and last one */
  for(i=0; i<nodesN; i++) {
    lab = treeOrderN[i]; /* label of node */
    pos = treePosN[lab]; /* position of node in treeN */
    if(treeN[pos].m > 0) {
      labFirst = lab; /* hold label of node */
      posFirst = i;   /* in preorde */
      break;
    }
  }

  printf("FIRST: preorder: %d pInP: %d label: %d (%s) %d\n",posFirst,pos,lab,treeN[pos].o, labFirst);
  int path2root1[N_NODES],steps1=0;
  j = treePosN[labFirst];
  while(treeN[j].p) {
    /* printf("treeN[%3d].n = %d\n",j,treeN[j].n); */
    path2root1[steps1] = treeN[j].n;
    j = treePosN[(treeN[j].p)->n];
    steps1++;
  }
  path2root1[steps1] = treeN[j].n;
  //for(i=0; i<=steps1; i++) { printf("%d,",path2root1[i]); } printf("\n");


  for(i=nodesN-1; i>=posFirst; i--) {
    lab = treeOrderN[i]; /* label of node */
    pos = treePosN[lab]; /* position of node in treeN */
    if(treeN[pos].m > 0) {
      labLast = lab; 
      posLast = i; /* in preorder */
      break;
    }
  }

  printf("LAST:  preorder: %d pInP: %d label: %d (%s) %d\n",posLast,pos,lab,treeN[pos].o, labLast);
  /* no mirs in tree ! species probably not in tree?! and this is the only species in the input file!!!*/

  if(posFirst > posLast) {
    return -1;
  }
  for(i=posFirst; i<=posLast; i++) {
    if(treeN[i].m > *kMax) *kMax = treeN[i].m;
    if(treeN[i].m < *kMin) *kMin = treeN[i].m;
  }

  int path2root2[N_NODES],steps2=0;
  j = treePosN[labLast];
  while(treeN[j].p) {
    path2root2[steps2] = treeN[j].n;
    j = treePosN[(treeN[j].p)->n];
    steps2++;
  }
  path2root2[steps2] = treeN[j].n;
  //for(i=0; i<=steps2; i++) { printf("%d,",path2root2[i]); } printf("\n");

  /* int maxsteps; */
  /* if(steps1 > steps2) { maxsteps = steps1; }  */
  /* else { maxsteps = steps2; } */

  if(posLast == posFirst) { lca = labFirst; }
  else {
    printf("kMax: %d kMin: %d\n", *kMax, *kMin); 
    
    i = steps1; j = steps2;
    while(path2root1[i] == path2root2[j]) { i--; j--; }
    i++; j++;
    //printf("i: %d j: %d lca: %d (%s)\n",i,j,path2root1[i],treeN[treePosN[path2root1[i]]].o);
    
    
    lca = path2root1[i]; /*  label of LCA node! */
  }
  
  int lca_pos = treePosN[lca];
  int lca_ord = treeN[lca_pos].pInP;
  int posX;
    
  for(j=0; j<lca_ord; j++) {
    lab = treeOrderN[j];
    posX = treePosN[lab];
    treeN[posX].seenBW = 1;
    treeN[posX].m = 0;
    /* printf("B: treeN[%d].o = %s .. initialized\n",posX,treeN[posX].o); */
  }
  printf("--\n");

  /* find rightmost first pos in preorder, not element of subtree! */
  /* start in lca */

  //  int lastleafpos = lca_pos; /* pos in treeN */
  int max,maxI,curpos;

  curpos = lca_pos;
    
  while(treeN[curpos].nc > 0) {
    lab = treeN[curpos].c[0];
    pos = treePosN[lab];
    max = treeN[pos].pInP;
    maxI = pos;
    for(i=1;i<treeN[curpos].nc;i++) {
      lab = treeN[curpos].c[i];
      pos = treePosN[lab];
      if(treeN[pos].pInP > max) {
	max = treeN[pos].pInP;
	maxI = pos;
      }
    }
    curpos = maxI;
  }

  /* printf("maxI: %d (%s)\n", curpos, treeN[curpos].o); */

  for(j=treeN[curpos].pInP+1; j<nodesN; j++) {
  // for(j=lastleafpos+1; j<nodesN; j++) {
    lab = treeOrderN[j];
    posX = treePosN[lab];
    treeN[posX].seenBW = 1;
    treeN[posX].m = 0;
    /* printf("A: treeN[%d].o = %s .. initialized\n",posX,treeN[posX].o); */
  }
  
  
  return lca;
}


/* -------------------------------------------------------------------------- */

/* prints the treeN in all its elements as POSTSCRIPT */
void gl_printTreePSN(int lca, char * fname, char * type) {

  int i, t, maxL=21, N=0, n_leaves=0, c, lca_pos;
  float x, y, w, h, x1, y1, x2, y2;
  FILE *T;
  //  float xn[N_NODES], yn[N_NODES];
  float *xn, *yn;

  //  printf("treeN[%d].m = %d\n",331,treeN[331].m);
  printf("type: %s\n",type);
  maxL = 0;
  for(i=0; i<nodesN; i++) {
    if(treeN[i].nc == 0) { n_leaves++; }
    if(treeN[i].level > maxL) { maxL = treeN[i].level; }
  }
  xn = (float *) calloc(nodesN+1, sizeof(float));
  yn = (float *) calloc(nodesN+1, sizeof(float));

  //  for(i=0; i<N_NODES; i++) { xn[i] = -1; yn[i] = -1; }
  for(i=0; i<nodesN; i++) { xn[i] = -1; yn[i] = -1; }

  lca_pos = treePosN[lca];
  /* width of level (w = 20) */
  w = 490 / (maxL+1);
  /* height of level (h = 9) */
  //h = 780 / N_LEAVES;  
  h = 820 / n_leaves;

  /* adjust fontsize of inner nodes to fit between the edges */
  int fontsize = h*0.8;
  if(fontsize > 10) fontsize = 10;

  T = fopen(fname, "w");

  fprintf(T, "%%!PS-Adobe-2.1 EPSF-2.0\n");
  fprintf(T, "%%"); fprintf(T, "%%Title: Efficient Prediction of Paralog Evolution - %s\n",fname);
  fprintf(T, "%%"); fprintf(T, "%%Creator: ePoPE 1.0 -- by Jana Hertel\n");
  fprintf(T, "%%"); fprintf(T, "%%DocumentFonts: Helvetica\n");
  fprintf(T, "%%"); fprintf(T, "%%BoundingBox: 0 0 595 842\n");
  fprintf(T, "%%"); fprintf(T, "%%EndComments\n\n");

  /* start after Metazoa */
  x = 45; 
  y = 820;

  fprintf(T, "1 setlinecap\n");
  
  /* get nodes by tree preorder */
  //  for(i=0; i<329; i++) {
  //  for(i=0; i<nodesN; i++) {
  int drawnNodes = 0;
  i = 0;
  while(drawnNodes < nodesN) {
    t = treePosN[treeOrderN[i]];
    //    printf("treeN[%d].o = %s\n",t,treeN[t].o);
    if(xn[t] != -1 && treeN[t].p !=NULL) {
      i = (treeN[t].p)->pInP;
      goto nextWhile3;
    }
    /* draw the horizontal lines of the leafs */
    //    if(tree[t].l == NULL && tree[t].r == NULL) {
    if(treeN[t].nc == 0) { /* this is a leaf */
      /* x1 = x + ((tree[t].level) * w); */ /* all leaves end at their level */
      x1 = 505; /* all leaves end at 505 */
      y1 = y - (N * h) + 1;
      fprintf(T, "gsave newpath\n");
      fprintf(T, "%.2f %.2f moveto %.2f %.2f lineto\n",x1,y1,x1+2,y1);
      fprintf(T, "%.2f %.2f lineto\n",x1+2,y1-2);
      fprintf(T, "%.2f %.2f lineto\n",x1,y1-2);
      fprintf(T, "%.2f %.2f lineto\n",x1,y1);
      fprintf(T, "closepath\n fill stroke\n");
      fprintf(T, "grestore\n");
      
      fprintf(T, "/Helvetica findfont 7 scalefont setfont\n");
      fprintf(T, "%.2f %.2f moveto (%s) show\n",545.,y1-3,treeN[t].o);
      /* node label */
      fprintf(T, "/Helvetica findfont 6 scalefont setfont\n");
	fprintf(T, "%.2f %.2f moveto (%-4d) show\n\n",530.,y1-3,treeN[t].n);
      /* node score (No. of genes) */

      fprintf(T, "gsave /Helvetica findfont 6 scalefont setfont\n");
      /*if(strcmp(type,"none") == 0) {
	fprintf(T, "0 0.4 0 setrgbcolor %.2f %.2f moveto (%4d) show\n",507.,y1-3,treeN[t].m);
	} else {*/
	fprintf(T, "0 0.4 0 setrgbcolor %.2f %.2f moveto (%4d) show\n",505.,y1-3,treeN[t].m);
	/*}*/
      fprintf(T, "grestore\n");
      
      /* leaf gain/loss */
      fprintf(T, "gsave /Helvetica findfont 6 scalefont setfont\n");
      /*if(strcmp(type,"none") != 0) {*/
	fprintf(T, "0 0 1 setrgbcolor %.2f %.2f moveto (%2d) show\n",515.,y1-3,treeN[t].gain);
	fprintf(T, "1 0 0 setrgbcolor %.2f %.2f moveto (%2d) show\n",520.,y1-3,treeN[t].loss);
	fprintf(T, "grestore\n");
	/*      }*/
      
      drawnNodes++;
      xn[t] = x1; yn[t] = y1;
      N++;
      i = (treeN[t].p)->pInP;
      goto nextWhile3;

    } else { /* inner node */
      /* check if all children have coords */
      for(c=0; c<treeN[t].nc; c++) { /* if not step to uncoordinated children recursively */
	if(xn[treePosN[treeN[t].c[c]]] == -1) { 
	  i = treeN[treePosN[treeN[t].c[c]]].pInP;
	  goto nextWhile3;
	}
      }
      /* all children coordinated */
      /* yn[t] = yn[treeN[treePosN[treeN[t].c[0]]].pInP]; */
      /* float yChildren=yn[treePosN[treeN[t].c[treeN[t].nc-1]]]; */
      float yChildren = 0;
      for(c=0; c<treeN[t].nc; c++) {
	yChildren += yn[treePosN[treeN[t].c[c]]];
      }
      yChildren /= treeN[t].nc;
      yn[t] = yChildren;
      xn[t] = x + (treeN[t].level*w);
      x1 = xn[t]; y1 = yn[t];
      if(t == lca_pos) {
	fprintf(T, "gsave\n");
	fprintf(T, "%.2f %.2f moveto %.2f %.2f lineto\n",x1,y1-1,x1,y1+3);
	fprintf(T, "%.2f %.2f lineto %.2f %.2f lineto\n",x1-5,y1-1,x1,y1-5);
	fprintf(T, "%.2f %.2f lineto\n",x1,y1-1);
	fprintf(T, "fill closepath\nstroke\n");
	fprintf(T, "grestore\n");
      }
      
      /* node label */
      fprintf(T, "/Helvetica findfont %d scalefont setfont\n", fontsize);
      fprintf(T, "%.2f %.2f moveto (%s) show\n",x1-20,y1+1,treeN[t].o);
    
      /* node score */
      if(type == NULL) {
	fprintf(T, "gsave /Helvetica findfont %d scalefont setfont\n", fontsize);
	fprintf(T, "0 0.4 0 setrgbcolor %.2f %.2f moveto (%d) show\n",x1+2,y1-7,treeN[t].m);
	fprintf(T, "grestore\n");
      } else {
	if(strcmp(type, "all") == 0 || strcmp(type, "genes") == 0) {
	  /* if(treeN[t].m > 0) { */
	    fprintf(T, "gsave /Helvetica findfont %d scalefont setfont\n", fontsize);
	    fprintf(T, "0 0.4 0 setrgbcolor %.2f %.2f moveto (%d) show\n",x1+2,y1-7,treeN[t].m);
	    fprintf(T, "grestore\n");
	  /* } */
	} 
      }

      /* gain */
      if(type == NULL) {
	if(treeN[t].gain > 0) {
	  fprintf(T, "/Helvetica findfont %d scalefont setfont\n", fontsize);
	  fprintf(T, "gsave 0 0 1 setrgbcolor\n");
	  fprintf(T, "%.2f %.2f moveto (\\(+%d\\)) show\n",x1+7,y1-7,treeN[t].gain);
	  fprintf(T, "grestore\n");
	}
      } else {
	if(strcmp(type, "all") == 0) {
	  if(treeN[t].gain > 0) {
	    fprintf(T, "/Helvetica findfont %d scalefont setfont\n", fontsize);
	    fprintf(T, "gsave 0 0 1 setrgbcolor\n");
	    fprintf(T, "%.2f %.2f moveto (\\(+%d\\)) show\n",x1+7,y1-7,treeN[t].gain);
	    fprintf(T, "grestore\n");
	  }
	} else if(strcmp(type, "gain") == 0) {
	  if(treeN[t].gain > 0) {
	    fprintf(T, "/Helvetica findfont %d scalefont setfont\n", fontsize);
	    fprintf(T, "gsave 0 0 1 setrgbcolor\n");
	    fprintf(T, "%.2f %.2f moveto (\\(+%d\\)) show\n",x1+2,y1-7,treeN[t].gain);
	    fprintf(T, "grestore\n");
	  }
	} else if(strcmp(type, "gainFam") == 0) {
	  if(treeN[t].gain > 0) {
	    fprintf(T, "/Helvetica findfont %d scalefont setfont\n", fontsize);
	    fprintf(T, "gsave 0 0 1 setrgbcolor\n");
	    fprintf(T, "%.2f %.2f moveto (\\(+%d\\)) show\n",x1+2,y1-7,treeN[t].gainFam);
	    fprintf(T, "grestore\n");
	  }	
	}
      }

      /* loss */
      if(type == NULL) {
	if(treeN[t].loss > 0) {
	  fprintf(T, "/Helvetica findfont %d scalefont setfont\n", fontsize);
	  fprintf(T, "gsave 1 0 0 setrgbcolor\n");
	  fprintf(T, "%.2f %.2f moveto (\\(-%d\\)) show\n",x1+12,y1-7,treeN[t].loss);
	  fprintf(T, "grestore\n");
	}
      } else { 
	if(strcmp(type, "all") == 0) {
	  if(treeN[t].loss > 0) {
	    fprintf(T, "/Helvetica findfont %d scalefont setfont\n", fontsize);
	    fprintf(T, "gsave 1 0 0 setrgbcolor\n");
	    fprintf(T, "%.2f %.2f moveto (\\(-%d\\)) show\n",x1+12,y1-7,treeN[t].loss);
	    fprintf(T, "grestore\n");
	  }
	} else if(strcmp(type, "loss") == 0) {
	  if(treeN[t].loss > 0) {
	    fprintf(T, "/Helvetica findfont %d scalefont setfont\n", fontsize);
	    fprintf(T, "gsave 1 0 0 setrgbcolor\n");
	    fprintf(T, "%.2f %.2f moveto (\\(-%d\\)) show\n",x1+2,y1-7,treeN[t].loss);
	    fprintf(T, "grestore\n");
	  }
	} else if(strcmp(type, "lossFam") == 0) {
	  if(treeN[t].loss > 0) {
	    fprintf(T, "/Helvetica findfont %d scalefont setfont\n", fontsize);
	    fprintf(T, "gsave 1 0 0 setrgbcolor\n");
	    fprintf(T, "%.2f %.2f moveto (\\(-%d\\)) show\n",x1+2,y1-7,treeN[t].lossFam);
	    fprintf(T, "grestore\n");
	  }
	}
      }
      
      /* stroke paths to all children */
      for(c=0; c<treeN[t].nc; c++) {
	x2 = xn[treePosN[treeN[t].c[c]]];
	y2 = yn[treePosN[treeN[t].c[c]]]-1;
	fprintf(T, "%.2f %.2f moveto %.2f %.2f lineto\n",x1,y1,x1,y2);
	fprintf(T, "%.2f %.2f moveto %.2f %.2f lineto stroke\n",x1,y2,x2,y2);
      }
      drawnNodes++;

      goto nextWhile3;

    nextWhile3:
      continue;
    }
  } 

  fprintf(T, "%.2f %.2f moveto %.2f %.2f lineto stroke\n\n",xn[0],yn[0],xn[0]-w,yn[0]);
  fprintf(T, "showpage\n");
  fflush(T);
  fclose(T);
  free(xn);
  free(yn);
  
}

/* -------------------------------------------------------------------------- */

void gl_printTreeN(char * fname) {

  int j, c;
  FILE *fp=NULL;
  
  if(fname == NULL) { /* write to stdout */

    for(j=0; j<nodesN; j++) {
      printf("pos: %3d lab: %3d kids: ",treeN[j].pInT,treeN[j].n);
      if(treeN[j].nc == 0) { printf("%d ", -1); }
      else {
	for(c=0; c<treeN[j].nc-1; c++) { printf("%d,", treeN[j].c[c]);}
	printf("%d ", treeN[j].c[(treeN[j].nc)-1]);
      }
      printf("pInP: %3d ",treeN[j].pInP);
      if(treeN[j].p == NULL) printf("par: -1 "); 
      else printf("par: %3d ",(treeN[j].p)->n);
      printf("mirs: %3d ",treeN[j].m);
      printf("gain: %3d loss %3d\t seenBW: %3d ",treeN[j].gain,treeN[j].loss,treeN[j].seenBW);
      printf("gainFam: %3d\t lossFam: %3d\t %s\n",treeN[j].gainFam,treeN[j].lossFam,treeN[j].o);
    }
    
  } else { /* write to file */

    fp = fopen(fname, "w");
    for(j=0; j<nodesN; j++) {
      fprintf(fp,"pos: %3d lab: %3d kids: ",treeN[j].pInT,treeN[j].n);
      if(treeN[j].nc == 0) { fprintf(fp, "%d ", -1); }
      else {
	for(c=0; c<treeN[j].nc-1; c++) { fprintf(fp, "%d,", treeN[j].c[c]);}
	fprintf(fp, "%d ", treeN[j].c[(treeN[j].nc)-1]);
      }
      fprintf(fp,"pInP: %3d ",treeN[j].pInP);
      if(treeN[j].p == NULL) fprintf(fp, "par: -1 "); 
      else fprintf(fp,"par: %3d ",(treeN[j].p)->n);
      fprintf(fp,"mirs: %3d ",treeN[j].m);
      fprintf(fp,"gain: %3d loss %3d\t seenBW: %3d ",treeN[j].gain,treeN[j].loss,treeN[j].seenBW);
      fprintf(fp,"gainFam: %3d\t lossFam: %3d\t %s\n",treeN[j].gainFam,treeN[j].lossFam,treeN[j].o);
      
    }
    fclose(fp);
    
  }

  return;
}

/* -------------------------------------------------------------------------- */

void gl_printS(float **S, int m, int n) {

  int i,j;

  printf("-   \t");
  for(j=0; j<n; j++){
    printf("  %c%c%c \t",treeN[j].o[0],treeN[j].o[1],treeN[j].o[2]);
  }
  printf("\n");

  printf("i\\j \t");
  for(j=0; j<n; j++)
    printf("%5d\t",j);
  printf("\n");

  //  printf("m: %d n: %d\n", m, n); exit(0)
  for(i=0; i<m; i++) {
    printf("%3d:\t",i);
    for(j=0; j<n; j++) {
      printf("%5.2f\t", S[i][j]);
    }
    printf("\n");
  }
  printf("\n");

  return;
  
}

/* -------------------------------------------------------------------------- */

