#include "read.h"

/* weight to gain a miRNA for each node */
float treeGainWeight[N_NODES];

/* weight to loose a miRNA for each node */
float treeLossWeight[N_NODES];

/* tree with max. N_CHILDREN number of child nodes */
struct gl_nodeN treeN[N_NODES];

/* preorder of treeN */
int treeOrderN[N_NODES];
/* position of nodes in treeN[] according to preorder */
int treePosN[N_NODES];

/* number of used nodes in built tree */
int nodesN;


/* -------------------------------------------------------------------------- */

struct gl_arguments gl_readArguments(int argc, char *argv[]) {

  int i;
  gl_arguments args;

  if(argc < 2) usage();
  /* printf("argc: %d, argv[0]: %s\n",argc,argv[0]); */
  args = gl_initArguments();

  for(i=0; i<argc; i++) {
    if(argv[i][0]=='-') {
      switch(argv[i][1]) {
      case 'h':
        usage();
        break;
      case 'v':
        version();
        break;
      case 'i':
	/* get alignment file */
	if(argc < i+2) usage();
	args.infile = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	strcpy(args.infile,argv[i+1]);
        break;
      case 't':
	/* get tree */
	if(argc < i+2) usage();
	args.treefile = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	strcpy(args.treefile, argv[i+1]);
	break;
      case 'w':
	/* get weight array */
	if(argc < i+2) usage();
	args.scores = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	strcpy(args.scores, argv[i+1]);
	break;
      case 'o':
	/* get dat-outfile name */
	if(argc < i+2) usage();
	args.outfile = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	strcpy(args.outfile, argv[i+1]);
	break;
      case 'p':
	/* get ps-outfile name */
	if(argc < i+2) usage();
	args.psfile = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	strcpy(args.psfile, argv[i+1]);
	break;
      case 'c':
	/* get a collection of ePoPE output created with ePoPE.summarize.pl and draw tree */
	if(argc < i+2) usage();
	args.collectfile = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	strcpy(args.collectfile, argv[i+1]);
	break;
      case 'l':
	/* activate loss-backtracing */
	args.lossBT = 1;
	break;
      case '-':
	if(strcmp(argv[i],      "--help"     ) ==0) usage();
        else if(strcmp(argv[i], "--version"  ) ==0) version();
	else if(strcmp(argv[i], "--clw"      ) ==0) args.clw = 1;
	else if(strcmp(argv[i], "--stk"      ) ==0) args.stk = 1;
	else if(strcmp(argv[i], "--type"     ) ==0) {
	  if(argc < i+2) usage();
	  args.type = (char *) calloc(strlen(argv[i+1]) + 1, sizeof(char));
	  strcpy(args.type, argv[i+1]);
	}
	else usage();
	break;
      default:
	usage();
      }
    }
  }

  if(args.infile == NULL) {
    if(args.collectfile == NULL) {
      gl_freeArguments(args);
      usage();
    }    
  }

  /* if infile is set, collectfile is ignored ! */
  if(args.infile != NULL && args.collectfile !=NULL) { 
    free(args.collectfile); 
    args.collectfile = NULL;
  }
  if(args.treefile == NULL) {
    gl_freeArguments(args);
    usage();
  }
  /* if(args.scores == NULL) { */
  /*   gl_freeArguments(args); */
  /*   usage(); */
  /* } */
  
  if(args.clw == 1 && args.stk == 1) { /* ?? -> check type of aln */
    args.clw = 0; args.stk = 0;
  }
    
  return args;
  
}

/* init functions ----------------------------------------------------------- */

struct gl_arguments gl_initArguments() {

  gl_arguments a;

  a.infile = NULL;
  a.treefile = NULL;
  a.outfile = NULL;
  a.psfile = NULL;
  a.scores = NULL;
  a.collectfile = NULL;
  a.type = NULL;
  a.clw = 0;
  a.stk = 0;
  a.lossBT = 0;
  
  return a;
  
}

/* print functions ---------------------------------------------------------- */

void gl_printArguments(gl_arguments ga) {

  fprintf(stdout, "\nArguments:\n----------\n");
  fprintf(stdout, "Input aln file:       %s\n",ga.infile);
  fprintf(stdout, "Input tree file:      %s\n",ga.treefile);
  if(ga.outfile != NULL) fprintf(stdout, "Output file:   %s\n",ga.outfile);
  else fprintf(stdout, "Data tree file:   -\n");
  if(ga.psfile != NULL) fprintf(stdout, "{PS-output file:   %s\n",ga.psfile);
  else fprintf(stdout, "PS-output file:   -\n");
  fprintf(stdout, "Score array file:  %s\n",ga.scores);
  if(ga.type != NULL) fprintf(stdout, "Type: %s\n", ga.type);
  else fprintf(stdout, "Type: -\n");
}

/* -------------------------------------------------------------------------- */

void usage() {

  fprintf(stdout, "\n+--------------------------------------------------+");
  fprintf(stdout, "\n| ePoPE 1.0                                        |");
  fprintf(stdout, "\n|                                                  |");
  fprintf(stdout, "\n| ePoPE - efficent Prediction of Paralog Evolution |");
  fprintf(stdout, "\n+==================================================+\n");
  fprintf(stdout, "\nePoPE predicts a maximal parsimony solution of gain and loss events of a gene family with paralogs.\n");

  fprintf(stdout,"\nUsage: ePoPE [ arguments ] -i ALNFILE -t TREEFILE\n\n");
  fprintf(stdout, "arguments: [-w WEIGHTFILE]\n");
  fprintf(stdout, "           [-o OUTFILE] [-p PS-OUTFILE]\n");
  fprintf(stdout, "           [-c COLLECTFILE]\n");
  fprintf(stdout, "           [-h,--help] [-v,--version]\n");
  fprintf(stdout, "           [--type TYPE]\n\n");

  fprintf(stdout, "%-20s Input alignment FILE in CLUSTALW/STOCKHOLM format. [REQUIRED]\n", "-i FILE");
  fprintf(stdout, "%-20s Input tree FILE see example.tree.dat format. [REQUIRED]\n", "-t FILE");
  fprintf(stdout, "%-20s Input weight array FILE. [REQUIRED]\n\n", "-w FILE");
  fprintf(stdout, "%-20s Output FILE for tree data. Default is 'INFILE.dat'.[OPTIONAL]\n\n", "-o FILE");
  fprintf(stdout, "%-20s Output FILE for PS-tree data. Default is 'INFILE.ps'. [OPTIONAL]\n\n", "-p FILE");
  fprintf(stdout, "%-20s FILE is a collection of calls to ePoPE with the same tree on a set of gene families created via 'ePoPE.summarize.pl'. This option forces ePoPE to draw this summarized tree. You must provide the tree file you used for the single ePoPE calls with -t option. [OPTIONAL]\n\n", "-c FILE");
  fprintf(stdout, "%-20s TYPE is one of {genes, gainFam, lossFam, gain, loss, all}. Is the type of values that are plotted in the tree. Default: 'all'. [OPTIONAL]\n\n", "-c FILE");
  fprintf(stdout, "%-20s Show this help message.\n", "-h,--help");
  fprintf(stdout, "%-20s Show version information.\n", "-v,--version");

  fprintf(stdout, "\nExample call:\n\n");
  fprintf(stdout, "\t./ePoPE -i example/example.stk  -t example/example.tree.dat -p example/example.ePoPE.out.ps -o example/example.ePoPE.out --type all\n\n");
  
  fprintf(stdout, "\nPlease feel free to contact me for comments, bug-reports, etc.\n\n");

  fprintf(stdout, "--\n");
  version();

}

/* -------------------------------------------------------------------------- */

void version() {
  
  fprintf(stdout, "ePoPE 1.0\n\n");
  fprintf(stdout, "Auhthor: Jana Hertel:\n");
  fprintf(stdout, "         jana@bioinf.uni-leipzig.de\n\n");
  fprintf(stdout, "Date:    November, 2014\n\n");
  
  exit(EXIT_SUCCESS);
  
}

/* free functions ----------------------------------------------------------- */

void gl_freeArguments(gl_arguments ga) {

  if(ga.infile != NULL)   free(ga.infile);
  if(ga.treefile != NULL) free(ga.treefile);
  if(ga.scores != NULL)   free(ga.scores);
  if(ga.outfile != NULL)  free(ga.outfile);
  if(ga.psfile != NULL)   free(ga.psfile);
  if(ga.collectfile != NULL)   free(ga.collectfile);
  if(ga.type != NULL)   free(ga.type);
}

/* -------------------------------------------------------------------------- */

void gl_readDATA(struct gl_arguments ga){

  int i=0, read;
  FILE * fp;
  char line[2000],* name, abbr[4];
  float g=-1, l=-1;

  /* read tree, build datastructure gl_node tree[] */
  gl_buildTree(ga.treefile);

  if(ga.collectfile !=NULL) { 
    gl_readSummaryTreeN(ga.collectfile); 
    return; 
  }
  
  /* read alignment */
  fp = fopen(ga.infile, "r");

  if(fp == NULL) {
    fprintf(stderr, "Cannot access input file: '%s'.\n", ga.infile);
    gl_freeArguments(ga);
    exit(EXIT_FAILURE);
  }

  /* read file linewise */
  char *pch;
  while(fgets(line, 2000, fp)) { 
    /* extract first word */
    pch = strchr(line, ' ');
    if(pch == NULL) continue;
    name = (char *) calloc(pch-line+2, sizeof(char));
    strncpy(name, line, pch-line+1);
    //    read = sscanf(line, "%s %s", name, seq); 
    //if(read == -1) continue;
    
    if(strcmp(name, "STOCKHOLM")     ==0) { free(name); continue;}
    else if(strcmp(name, "CLUSTAL")  ==0) { free(name); continue;}
    else if(strcmp(name, "CLUSTALW") ==0) { free(name); continue;}
    else if(strcmp(name, "//")       ==0) { free(name); continue;}
    else if(strlen(name) == 0) { free(name); continue;}
    /* exclude lines starting with '#' (comments) */
  if(name[0] == 35) { free(name);continue; } 
  //printf("line: %s\nname1: %s\n",line,name);
    
    /* extract species abbreviation */
    int namesplit_n = 0;
    char **namesplit = gl_splitString(name, ':', &namesplit_n);
    /* printf("namesplit_n = %d -> %s\n",namesplit_n, name); */
    if(namesplit_n == 1) {
      //      if(strncpy(abbr, name, 3)) {
      if(strncpy(abbr, name, 3)) {
	abbr[3] = '\0';
	//printf("abbr: %s\n",abbr);
	for(i=0; i<nodesN; i++) {
	  /* printf ("%d %s %s\n", i,treeN[i].o,abbr); */
	  if(strcmp(treeN[i].o, abbr) == 0) {
	    treeN[i].m++;
	    treeN[i].P++;
	    break;
	  }
	  /* if(strcmp(tree[i].o, abbr) == 0) { */
	  /*   tree[i].m++; */
	  /*   break; */
	  /* } */
	}
      }
    } else {
      if(strncpy(abbr,namesplit[1],3))  {
	abbr[3] = '\0';

	for(i=0; i<N_NODES; i++) {
	  /* printf ("%d %s %s\n", i,tree[i].o,abbr); */
	  if(strcmp(treeN[i].o, abbr) == 0) {
	    treeN[i].m++;
	    treeN[i].P++;
	    break;
	  }
	  /* if(strcmp(tree[i].o, abbr) == 0) { */
	  /*   tree[i].m++; */
	  /*   break; */
	  /* } */
	}
      }
      
    }
    int z;
    for(z=0; z<namesplit_n; z++) { free(namesplit[z]); }
    free(namesplit);
    free(name);

    abbr[0] = '\0';

   }

  fclose(fp);
  

  /* calculate P_v (max No of genes in subtree rooted at v) */
  gl_setPatInnerNodes();

  
  /* for(i=0; i<nodesN; i++) { */
  /*    printf("treeN[%d].m = %d\t",i,treeN[i].m); */
  /*    printf("treeN[%d].P = %d\t",i,treeN[i].P); */
  /*    printf("treeN[%d].o = %s\n",i,treeN[i].o); */
  /* } */
  /* exit(0); */

  /* read weights for gain and loss */
  if(ga.scores !=NULL) {
    fp = fopen(ga.scores, "r");

    while(fgets(line, 512, fp)) {
      read = sscanf(line, "%d %f %f\n", &i, &g, &l);
      if(read == 0) continue;
      
      treeGainWeight[i] = g;
      treeLossWeight[i] = l;
    }
  
    fclose(fp);
  } else {
    for(i=0;i<nodesN;i++) {
      treeGainWeight[i] = 1.;
      treeLossWeight[i] = 1.;      
    }
  }
  
  return;

}

/* -------------------------------------------------------------------------- */

void gl_buildTree(char *filename){

  FILE *fp=NULL;
  char line[512];
  int i,n,n2, par=-1, read=0;
  int nodes=0,lines=0;
  char ch, **splitLine, **kids;

  
  fp = fopen(filename, "r");
  while(!feof(fp)) {
    ch = fgetc(fp);
    if(ch == '\n') lines++;
  }

  if(lines > N_NODES) {
    fprintf(stderr, "ERROR: you have more than '%d' nodes in your treefile. I cannot handle this!\nSorry.\n",N_NODES);
    fclose(fp);
    exit(0);
  }
  
  rewind(fp);
  
  /* while(fgets(line, 512, fp)) { */
  while(fgets(line, 512, fp)) {
    splitLine = gl_splitString(line, ' ',&n);
    strcpy(treeN[nodes].o,splitLine[0]);
    read = sscanf(splitLine[2], "%d", &treeN[nodes].n);    assert(read);
    read = sscanf(splitLine[4], "%d", &par);    assert(read);
    if(par == -1) { 
      treeN[nodes].parentLab = -1; 
      treeN[nodes].p = NULL; 
    } else treeN[nodes].parentLab = par;
    read = sscanf(splitLine[8], "%d", &treeN[nodes].level); assert(read);
    read = sscanf(splitLine[10], "%f", &treeN[nodes].gainW);  assert(read);
    read = sscanf(splitLine[12], "%f", &treeN[nodes].lossW);  assert(read);
    
    kids = gl_splitString(splitLine[6], ',', &n2);
    for(i=0; i<n2; i++) {
      read = sscanf(kids[i], "%d", &treeN[nodes].c[i]); assert(read);
      /* printf("kids[%d]=%s\n",i,kids[i]); */
    }

    treeN[nodes].m = 0;
    treeN[nodes].P = 0;
    if(n2 == 1 && treeN[nodes].c[0] == -1) {
      treeN[nodes].nc = 0;
      /* treeN[nodes].seenBW = 1; */
    } else {
      treeN[nodes].nc = n2;
      /* treeN[nodes].seenBW = -1; */
    }
    treeN[nodes].score = 0.;
    treeN[nodes].gain = 0;
    treeN[nodes].loss = 0;
    treeN[nodes].pInT = nodes;
    treeN[nodes].seenFW = -1;
    treeN[nodes].seenBW = -1;
     
    /* printf("tree[%d].n=%3d\tKIDS: %d tree[%d].o=%s\n",nodes,treeN[nodes].n,treeN[nodes].nc,nodes,treeN[nodes].o); */
    /* for(i=0; i<n2; i++) { printf("treeN[%d].c[%d] = %d\n",nodes,i,treeN[nodes].c[i]);    } */
    /* fflush(stdout); */
    nodes++;
    for(i=0; i<n; i++) { free(splitLine[i]); }
    free(splitLine);
    for(i=0; i<n2; i++) { free(kids[i]); }
    free(kids);
  }

  fclose(fp);

  nodesN = nodes;
  gl_label2pos();
  //  for(i=0;i<nodesN;i++) { printf("treePosN[%d]=%d\n",i,treePosN[i]); }
  /* printf("nodesN: %d\n",nodesN); */
  /* initialize treeOrderN */
  for(i=0; i<nodesN; i++) {
    treeOrderN[i] = -1;
  }

  int root=0, p=0;
  for(i=0; i<nodesN; i++) {
    if(treeN[i].parentLab == -1) { 
      root = i; 
      break; 
    }
  }

  p = gl_preorder(treeN[root], p);
  //  printf("nodesN: %d\nTree in pre-order:\n",nodesN);
  
  for(i=0;i<nodesN;i++) {
    p = treePosN[treeOrderN[i]];
    if(treeN[p].parentLab >=0 ) {
      int plab = treeN[p].parentLab;
      int pposintree = treePosN[plab];
      treeN[p].p = &treeN[pposintree];
    } else { treeN[p].p = NULL; }
    /* printf("treeOrderN[%3d] = %d -> pos in tree: %4d (%s) ",i,treeOrderN[i], p, treeN[p].o); */
    /* if(treeN[p].p) { printf(" parent: %d (%s)\n",(treeN[p].p)->n,(treeN[p].p)->o);} */
    /* else { printf("at root\n"); } */
  }
  
  //  exit(0);

}

/* -------------------------------------------------------------------------- */

/* fills array of treePosN with positions of labels */
/* index: label, value: position */
void gl_label2pos() {
  int i, j;

  for(i=0; i<nodesN; i++) {
    j = treeN[i].n;
    //    printf("T: treeN[%d].n=%-5d\t",i,treeN[i].n);
    treePosN[j] = i;
    //    printf("treePosN[%d]=%d\n",j,i);
  }

  return;
}

/* -------------------------------------------------------------------------- */

int gl_preorder(struct gl_nodeN node, int o) {

  int k, i;
  /* o is index in pre-order, value: label of node in treeN */
  treeOrderN[o] = node.n;
  treeN[treePosN[node.n]].pInP = o; /* stores position in preorder in corresp. node */
  
  //  treePosN[node.n] = node.pInT;

  /* printf("order[%d] = %d\t",o,treeOrderN[o]); */
  /* printf("  pos[%d] = %d\t",treeOrderN[o],treePosN[treeOrderN[o]]); */
  /* printf("%s -- %s\n", node.o, treeN[treePosN[node.n]].o); */
  
  o = o + 1;

  if(node.c[0] == -1) return o;/* leaf reached */

  for(k=0; k<node.nc; k++) {
    for(i=0; i<nodesN; i++) {
      if(treeN[i].n == node.c[k]) {
	o = gl_preorder(treeN[i], o);
	break;
      }
    }
  }

  return o;
}

/* -------------------------------------------------------------------------- */

  char** gl_splitString(char * str2split, const char a_delim, int *counter) {

  char ** result = 0;
  size_t count = 0;
  char * tmp = str2split;
  char * last_comma = 0;
  char delim[2];
  delim[0] = a_delim; delim[1] = 0;

      /* Count how many elements will be extracted. */
    while (*tmp)
    {
        if (a_delim == *tmp)
        {
            count++;
            last_comma = tmp;
        }
        tmp++;
    }

    /* Add space for trailing token. */
    count += last_comma < (str2split + strlen(str2split) - 1);

    /* Add space for terminating null string so caller
       knows where the list of returned strings ends. */
    count++;

    result = malloc(sizeof(char*) * count);

    if (result)
    {
        size_t idx  = 0;
        char* token = strtok(str2split, delim);

        while (token)
        {
            assert(idx < count);
            *(result + idx++) = strdup(token);
            token = strtok(0, delim);
        }
        assert(idx == count - 1);
        *(result + idx) = 0;
	if(token != NULL) free(token);
    }

    * counter = count-1;

    return result;
  
}

/* -------------------------------------------------------------------------- */
   /* calculate P_v (max No of genes in subtree rooted at v) */
void gl_setPatInnerNodes() {
  
  int i, lab;

  for(i=0; i<nodesN; i++) {
    lab = treePosN[treeOrderN[i]];

    /* printf("%d: treeOrderN[%d]=",i,i); */
    /* printf("%d (pos %d)",treeOrderN[i],lab); */
    /* printf("P=%d\n",treeN[lab].P); */
    
    treeN[lab].P = gl_getPv(lab);
  }

  return;
}

int gl_getPv(int label_v) {
  
  int u, maxP, curP;
  /* printf("treeN[%d].nc = %d\n",label_v, treeN[label_v].nc); */
  if(treeN[label_v].nc == 0) {
    /* printf("\t\tLEAVE: P=%d (m=%d)\n",treeN[label_v].P,treeN[label_v].m); */
    return treeN[label_v].P;
  } else {
    maxP = 0;
    for(u=0; u<treeN[label_v].nc; u++) {
      /* printf("\tu: %d\n",u); */
      curP = gl_getPv(treePosN[treeN[label_v].c[u]]);
      if(curP > maxP) { 
	maxP = curP; 
      }
    }
    return maxP;
  }

}

/* -------------------------------------------------------------------------- */
/* reads the summarized tree and stores values in treeN */
void gl_readSummaryTreeN(char * file) {
  FILE *fp=NULL;
  char line[512];
  int j,i,n,n2, par=-1, read=0;
  int lines=0;
  char ch, **splitLine, **kids;

  
  fp = fopen(file, "r");
  while(!feof(fp)) {
    ch = fgetc(fp);
    if(ch == '\n') lines++;
  }

  if(lines > N_NODES) {
    fprintf(stderr, "ERROR: you have more than '%d' nodes in your treefile. I cannot handle this!\nSorry.\n",N_NODES);
    fclose(fp);
    exit(0);
  }
  
  rewind(fp);
  
  /* while(fgets(line, 512, fp)) { */
  j = 0;
  while(fgets(line, 512, fp)) {
    splitLine = gl_splitString(line, ' ',&n);
    read = sscanf(splitLine[1], "%d", &j); assert(read);
    read = sscanf(splitLine[3], "%d", &treeN[j].n); assert(read); 
    read = sscanf(splitLine[7], "%d", &treeN[j].pInP); assert(read);
    read = sscanf(splitLine[11], "%d", &treeN[j].m); assert(read);

    read = sscanf(splitLine[13], "%d", &treeN[j].gain); assert(read);
    read = sscanf(splitLine[15], "%d", &treeN[j].loss); assert(read);
    read = sscanf(splitLine[19], "%d", &treeN[j].gainFam); assert(read);
    read = sscanf(splitLine[21], "%d", &treeN[j].lossFam); assert(read);

    /* parent */
    read = sscanf(splitLine[9], "%d", &par);    assert(read);
    if(par == -1) { 
      treeN[j].parentLab = -1; 
      treeN[j].p = NULL; 
    } else treeN[j].parentLab = par;

    /* name */
    //treeN[j].o = (char *) calloc(strlen(splitLine[22]) + 1, sizeof(char));
    strcpy(treeN[j].o, splitLine[22]); 

    /* kids */
    kids = gl_splitString(splitLine[5], ',', &n2);
    for(i=0; i<n2; i++) {
      read = sscanf(kids[i], "%d", &treeN[j].c[i]); assert(read);
      /* printf("kids[%d]=%s\n",i,kids[i]); */
    }

    j++;
    free(splitLine);
  }

  fclose(fp);
  
  nodesN = j;
  gl_label2pos();
  //  for(i=0;i<nodesN;i++) { printf("treePosN[%d]=%d\n",i,treePosN[i]); }
  /* printf("nodesN: %d\n",nodesN); */
  /* initialize treeOrderN */
  for(i=0; i<nodesN; i++) {
    treeOrderN[i] = -1;
  }

  int root=0, p=0;
  for(i=0; i<nodesN; i++) {
    if(treeN[i].parentLab == -1) { 
      root = i; 
      break; 
    }
  }

  p = gl_preorder(treeN[root], p);
  //  printf("nodesN: %d\nTree in pre-order:\n",nodesN);
  
  for(i=0;i<nodesN;i++) {
    p = treePosN[treeOrderN[i]];
    if(treeN[p].parentLab >=0 ) {
      int plab = treeN[p].parentLab;
      int pposintree = treePosN[plab];
      treeN[p].p = &treeN[pposintree];
    } else { treeN[p].p = NULL; }
    /* printf("treeOrderN[%3d] = %d -> pos in tree: %4d (%s) ",i,treeOrderN[i], p, treeN[p].o); */
    /* if(treeN[p].p) { printf(" parent: %d (%s)\n",(treeN[p].p)->n,(treeN[p].p)->o);} */
    /* else { printf("at root\n"); } */
  }
 
}
